#ifndef H_FWM
#define H_FWM

#include "utils.h"
#include "layout.h"
#include "atomic.h"

#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>
#include <X11/Xlib.h>
#include <X11/Xproto.h>
#include <X11/Xutil.h>
#include <X11/XKBlib.h>


enum { CurNormal, CurResize, CurMove, CurLast };

typedef struct {
  Display* dpy;
  int screen;
  Window root;
  Cursor cursor[CurLast];
} xstate;

typedef struct {
  unsigned int click;
  unsigned int mask;
  unsigned int button;
  void (*func)(Client*);
} Button;

typedef struct {
  unsigned int mod;
  KeySym keysym;
  void (*func)(int);
  int arg;
} Key;

extern xstate XSTATE;
extern Monitor* MONITOR;

extern char* hook_unmanage;
extern char* hook_manage;

#define BUTTONMASK              (ButtonPressMask|ButtonReleaseMask)
#define MOUSEMASK               (BUTTONMASK|PointerMotionMask)
#define CLEANMASK(mask)         (mask & ~(numlockmask|LockMask))
#define LENGTH(X)               (sizeof X / sizeof X[0])

#define SHIFTKEY ShiftMask
#define CTRLKEY ControlMask
#define MODKEY Mod1Mask
#define LINKEY Mod4Mask


//internals (not for user interaction)
unsigned int xorg_keycodetomodmask(unsigned int keycode);
int xorg_restack();//reapply the window order
int xorg_recalc();//if geometry changes, recalc & reapply the sizes
void xorg_checkprotocols(Client* c);
void xorg_getsizehints(Client* c);//update hints
void xorg_getstrings(Client* c);
//void xorg_clearurgent(Client* c);
//void xorg_updatewmhints(Client* c);
void xorg_refreshnetwmhints(Client* c, int clearskip);//sync server state
void xorg_grabbuttons(Client* c, int focused);
void xorg_grabkeys();
int xorg_getrootptr(int* x, int* y);
void xorg_refreshclientlist();
void xorg_refreshfocus();
void xorg_acknowledgetype(Client* c);
void xorg_refreshclientdesktop(Client* c);
void xorg_refreshalldesktop();
void xorg_desktopsupdate();

void xorg_exec(const char* cmd, char** argv);

//functions which happen to windows as we know them
//add or remove clients to our structure
int xclient_manage(Client* c);
int xclient_unmanage(Client* c);
//movement, manipulation
int xclient_show(Client* c);
int xclient_hide(Client* c);
int xclient_focus(Client* c);//take focus in sense of events
int xclient_raise(Client* c);//bring everything to top, also layer
int xclient_activate(Client* c);
int xclient_moveresize(Client* c);//apply reconfiguration
int xclient_configure(Client* c, unsigned int mask);//selective reconf
int xclient_changelayer(Client* c, Layer* new);
int xclient_kill(Client* c);//kill (not unmanage, this happens later)
int xclient_murder(Client* c, int sig);
int xclient_stackrel(Client* c, int stackrel);//change stacking type
int xclient_floatrel(Client* c, int floatrel);//change floating type
//layer-based
int xlayer_raise(Layer* l);
int xlayer_bringfront(Layer* l);
int xlayer_bringback(Layer* l);
int xlayer_visible(Layer* l, int visible);//change layer visibility

int xlayer_applylayout(char* string, int force);
int xdesktop_setdesktop(Dsktp* d);

//events that call above functions

void eventfun_mousemove(Client* c);
void eventfun_mouseresize(Client* c);
void eventfun_mousefocus(Client* c);
void eventfun_scrollup(Client* c);
void eventfun_scrolldn(Client* c);

void event_grab_alttab(int type);
void event_mouseclick(XEvent* e);
void event_keypress(XEvent* e);
void event_configure_request(XEvent* e);
void event_configure_notify(XEvent* e);
void event_destroy_notify(XEvent* e);
void event_client_focusin(XEvent* e);
void event_client_focusout(XEvent* e);
void event_client_maprequest(XEvent* e);
void event_client_propertynotify(XEvent* e);
void event_client_message(XEvent* e);
void event_mappingnotify(XEvent* e);
void event_unmapnotify(XEvent* e);

//user interaction one-shot functions

void user_alttab(int arg);

void user_togglelayer(int i);
void user_sendtolayer(int i);

void user_sendtonexttile(int i);
void user_nextintile(int i);

void user_desktopnext(int i);//relative; preset masks!
void user_desktopswitch(int i);//absolute; preset masks!

void user_kill(int arg);
void user_killtransient(int arg);
void user_client_tog_flrel(int floatrel);
void user_client_tog_strel(int stackrel);
void user_quit(int arg);
void user_terminal(int arg);

//globals
int xorg_checkotherwm();
int xorg_setup();
int xorg_scan();
int xorg_run();
int xorg_dispose();

#endif
